/* max186.c

 * Copyright (C) 2002 Claudio Girardi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111, USA
 */


#include <stdio.h>
#include <unistd.h>		/* for usleep() and getuid() */
#include <asm/io.h>


#define ADC_POWER 0x30		/* power from P0 and SLCT_OUT */
#define ADC_SCLK  0x04		/* serial clock from D2 */
#define ADC_CS    0x10		/* chip select from D4 */
#define ADC_EOC   0x40		/* end-of-conversion to ACK */
#define ADC_DOUTMASK 0x80	/* serial data out thru BUSY */
#define ADC_DINMASK  0x08	/* serial data in thru D3 */
#define ADC_DELAY 10		/* delay between parallel port commands */

#define BIT0  (1 << 0)
#define BIT1  (1 << 1)
#define BIT2  (1 << 2)
#define BIT3  (1 << 3)
#define BIT4  (1 << 4)
#define BIT5  (1 << 5)
#define BIT6  (1 << 6)
#define BIT7  (1 << 7)


static unsigned short outbyte;	/* parallel port output data */
static int port_base;		/* printer port base address */


int init_hw(int base)
{
  uid_t real_uid;
  gid_t real_gid;

  /* checking I/O permissions */
  if (ioperm(base, 3, 1)) {
    /* can not access the parallel port */
    return -1;
  } else {
    port_base = base;
    /* OK, now dropping root privileges */
    real_uid = getuid();
    setuid(real_uid);
    real_gid = getgid();
    setgid(real_gid);
    return 0;
  }
}


void close_hw(void)
{
  /* nothing to do */
}


static void rwbyte(unsigned short dout, unsigned short *din)
{
  unsigned short rbyte = 0, rbit, outbit, i;
  long int j;

  for (j = 0; j < ADC_DELAY; j++);
  /* read and write 8 bits */
  rbyte = 0;
  for (i = 1; i <= 8; i++) {
    /* get MSB (D7) */
    outbit = (dout & 128) / 128;
    /* output the bit */
    outbyte |= (outbit * ADC_DINMASK);
    outb((unsigned char) outbyte, port_base);
    for (j = 0; j < ADC_DELAY; j++);
    /* then set SCLK */
    for (j = 0; j < ADC_DELAY; j++);
    outbyte |= ADC_SCLK;
    outb((unsigned char) outbyte, port_base);
    for (j = 0; j < ADC_DELAY; j++);

    /* clear SCLK */
    for (j = 0; j < ADC_DELAY; j++);
    outbyte &= ~ADC_SCLK;
    outb((unsigned char) outbyte, port_base);
    for (j = 0; j < ADC_DELAY; j++);
    /* clean outbyte */
    outbyte &= ~(outbit * ADC_DINMASK);

    /* read bit */
    for (j = 0; j < ADC_DELAY; j++);
    /* since the input bit is read thru the /BUSY line it should be inverted */
    rbit = ((~(inb(port_base + 1))) & ADC_DOUTMASK) / ADC_DOUTMASK;
    for (j = 0; j < ADC_DELAY; j++);
    rbyte = (rbyte << 1) + rbit;
    //printf("%i\t%i\n", outbit, rbit);

    dout <<= 1;
  }

  (*din) = rbyte;
}



float get_sample(int channel)
{
  unsigned short inbyte, adc_data;
  long int result;
  float value;

  result = 0;

  if (channel == -1) {		/* channel -1 means "read the ADC offset" */
    /* the MAX186 doesn't have internal offset reading */
    return 0.0;
  } else {
    /* select channel and start conversion */
    /* use differential and bipolar conversion mode */
    /* see the Channel Selection Table in the Data Sheet */
    adc_data = 0x82 + ((channel & 0x03) << 4) + (0x4 << 4);
  }

  /* before clocking in data put CS and SCLK low */
  outbyte &= ~(ADC_CS | ADC_SCLK);
  outb((unsigned char) outbyte, port_base);

  rwbyte(adc_data, &inbyte);
  //usleep(250000); /* wait for 0.25 sec, allowing the input to stabilize */

  /* wait for EOC */
  while ((inb(port_base + 1) & ADC_EOC) == 0);

  /* dummy output byte */
  /* outbyte = 0x00; */
  /* read the 8 most significant bits */
  rwbyte(0x00, &inbyte);
  result = inbyte;
  result = result << 4;

  /* read the 4 least significant bits */
  rwbyte(0x00, &inbyte);
  result += (inbyte >> 4);

  /* set CS */
  outbyte |= ADC_CS;
  outb((unsigned char) outbyte, port_base);
  /* set SCLK */
  outbyte |= ADC_SCLK;
  outb((unsigned char) outbyte, port_base);

#ifdef DEBUG
  printf("result = %f\n", result);
#endif

  /* check the sign bit */
  if ((result & 0x800) == 0x800) {
    /* convert the two's complement reading */
    result = -4096 + result;
  }
#ifdef DEBUG
  printf("Result = %li\n", result);
#endif

  /* input voltage */
  value = 2.048 * (result / 2048.0);

  return value;
}
